from datetime import datetime

from src.typegenie import authenticator, User, Event, EventType, Author

# Authentication
# USER_ACCESS_TOKEN = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1c2VyX2lkIjoibXktbmV3LXVzZXIiLCJkZXBsb3ltZW50X2lkIjoibXktbmV3LWRlcGxveW1lbnQiLCJhY2NvdW50X2lkIjoiS1VORE9TRSIsInNlcV9udW0iOjEsImV4cCI6MTYxNjcwOTYxOCwiaWF0IjoxNjE2NzA2MDE4fQ.J0vcawZBJVxcO28_ouxTwIOUSABgTbkPo5d3WSUa-xk'
USER_ACCESS_TOKEN = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1c2VyX2lkIjoibXktbmV3LXVzZXIiLCJkZXBsb3ltZW50X2lkIjoibXktbmV3LWRlcGxveW1lbnQiLCJhY2NvdW50X2lkIjoiS1VORE9TRSIsInNlcV9udW0iOjEsImV4cCI6MTYxODUxNzQ0OSwiaWF0IjoxNjE4NTEzODQ5fQ.JD3_6wWnHXqN2dTEeuAz29PE5uiM5IYH6Fq5BthF66Q'


DEPLOYMENT_ACCESS_TOKEN = None
ACCOUNT_USERNAME = None
ACCOUNT_PASSWORD = None

if USER_ACCESS_TOKEN is not None:
    authenticator.authenticate_user(token=USER_ACCESS_TOKEN)
elif DEPLOYMENT_ACCESS_TOKEN is not None:
    authenticator.authenticate_deployment(token=DEPLOYMENT_ACCESS_TOKEN)
    # Then you can fallback to higher level API automatically by running following command
    authenticator.enable_auto_fallback()
elif ACCOUNT_USERNAME is not None and ACCOUNT_PASSWORD is not None:
    authenticator.authenticate_account(username=ACCOUNT_USERNAME, password=ACCOUNT_PASSWORD)
    # Then you can fallback to higher level API automatically by running following command
    authenticator.enable_auto_fallback()
else:
    raise RuntimeError('You must either have a user/deployment access token or account credentials')

# Ignore this. This is to run the sandbox environment
authenticator.enable_sandbox()

# Furthermore, since the access token expires automatically after a while, you can enable token auto renew using
authenticator.enable_auto_renew()

# Assuming that the user with id `my-new-user` exists.
user_id = 'my-new-user'
deployment_id = 'my-new-deployment'

user = User.get(user_id=user_id, deployment_id=deployment_id)
print('User: ', user)

# Create a new session
session_id = user.create_session()
print('Session: ', session_id)

# Get predictions
events = [Event(author_id='lost-soul-visitor',
                value='What is love?',
                event=EventType.MESSAGE,
                timestamp=datetime.utcnow(),
                author=Author.USER),
          Event(author_id='my-new-user',  # Note this is an agent already added as user to deployment
                value="Oh baby, don't hurt me",
                event=EventType.MESSAGE,
                timestamp=datetime.utcnow(),  # This should be time at which the event happened
                author=Author.AGENT)
          ]

print(user.get_completions(session_id=session_id, events=events, query="Don"))



from src.typegenie import authenticator, Deployment, Dialogue, Event, EventType, Author
from datetime import datetime

# Assuming that the deployment with id `my-new-deployment` exists.
deployment_id = 'my-new-deployment'

# Authentication
DEPLOYMENT_ACCESS_TOKEN = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJkZXBsb3ltZW50X2lkIjoibXktbmV3LWRlcGxveW1lbnQiLCJhY2NvdW50X2lkIjoiS1VORE9TRSIsImV4cCI6MTYyMDIyNDgwMiwic2VxX251bSI6MSwiaWF0IjoxNjIwMjIxMjAyfQ.09uD4GnJW0RmkPMiDH-65xVYV2Bf7rFy4o3qC5uZyII'


ACCOUNT_USERNAME = None
ACCOUNT_PASSWORD = None

# Enabling sandbox environment. Ignore this!
authenticator.enable_sandbox()

if DEPLOYMENT_ACCESS_TOKEN is not None:
    authenticator.authenticate_deployment(token=DEPLOYMENT_ACCESS_TOKEN)
elif ACCOUNT_USERNAME is not None and ACCOUNT_PASSWORD is not None:
    authenticator.authenticate_account(username=ACCOUNT_USERNAME, password=ACCOUNT_PASSWORD)
    # Then you can fallback to higher level API automatically by running following command
    authenticator.enable_auto_fallback()
else:
    raise RuntimeError('You must either have a deployment access token or account credentials')


# Furthermore, since the access token expires automatically after a while, you can enable token auto renew using
authenticator.enable_auto_renew()

# Assuming that the deployment with id `my-new-deployment` exists.
deployment = Deployment.get(deployment_id=deployment_id)
print('Deployment:', deployment)

# Dialogue upload

# Create Dialogue

my_dialogue_1 = Dialogue(dialogue_id='my-dialogue-1', metadata={'title': "What is love?"})

my_dialogue_1.events.append(Event(author_id='lost-soul-visitor',
                                  value='What is love?',
                                  event=EventType.MESSAGE,
                                  timestamp=datetime.utcnow(),
                                  author=Author.USER))
my_dialogue_1.events.append(Event(author_id='my-new-user',  # Note this is an agent already added as user to deployment
                                  value="Oh baby, don't hurt me",
                                  event=EventType.MESSAGE,
                                  timestamp=datetime.utcnow(),  # This should be time at which the event happened
                                  author=Author.AGENT))
my_dialogue_1.events.append(Event(author_id='lost-soul-visitor',
                                  value="Don't hurt me, no more",
                                  event=EventType.MESSAGE,
                                  timestamp=datetime.utcnow(),
                                  author=Author.AGENT))

my_dialogue_2 = Dialogue(dialogue_id='my-dialogue-2', metadata={'Artist': "Ping Floyd"})

my_dialogue_2.events.append(Event(author_id='system',
                                  value='Jam session begins:',
                                  event=EventType.CONTEXTUAL,
                                  timestamp=datetime.utcnow(),
                                  author=Author.SYSTEM))
my_dialogue_2.events.append(Event(author_id='pink-floyd-fan',
                                  value='Where were you when I was burned and broken? And where were you when I was '
                                        'hurt and I was helpless?',
                                  event=EventType.MESSAGE,
                                  timestamp=datetime.utcnow(),
                                  author=Author.USER))
my_dialogue_2.events.append(Event(author_id='my-new-user',  # Note this is an agent already added as user to deployment
                                  value="While the days slipped by from my window watching, I was staring straight "
                                        "into the shining sun. 'Cause the things you say and the things you do "
                                        "surround me",
                                  event=EventType.MESSAGE,
                                  timestamp=datetime.utcnow(),
                                  author=Author.AGENT))
my_dialogue_2.events.append(Event(author_id='pink-floyd-fan',
                                  value="Dying to believe in what you heard",
                                  event=EventType.MESSAGE,
                                  timestamp=datetime.utcnow(),
                                  author=Author.AGENT))


# Get existing dataset
dataset_id = 'my-new-dataset'  # Assumes that this already exists
existing_dataset = deployment.datasets(dataset_id=dataset_id)
print('Existing Dataset:', existing_dataset)

# upload dialogues
existing_dataset.upload(dialogues=[my_dialogue_1, my_dialogue_2]*100000)


# download
download_links = existing_dataset.get_download_links()
print(download_links)
doSomething = 1



import time
from typing import List
from random import randint, choice
from .lib import User, Event

try:
    import colorama
    from colorama import Back as B, Fore as F, Style
    from prompt_toolkit import PromptSession
    from prompt_toolkit.completion import Completer, Completion
    from prompt_toolkit.document import Document
    from prompt_toolkit.application.current import get_app
except ImportError as e:
    raise Exception('Extra dependencies are required to use CLI tool. Use `pip install typegenie[with-cli]`').with_traceback(e.__traceback__)

colorama.init()


def color(text, fore=F.RESET, back=B.RESET, reset=True):
    if reset:
        return f'{fore}{back}{text}{Style.RESET_ALL}'
    else:
        return f'{fore}{back}{text}'


def printc(text, fore=F.WHITE, back=B.RESET, reset=True):
    print(color(text, fore, back, reset))


def text_changed(buffer):
    buffer.start_completion(select_first=False)


def pre_run():
    app = get_app()
    buff = app.current_buffer
    buff.start_completion(select_first=False)
    # print('In the test!')


class TypeGenieCompleter(Completer):
    def __init__(self, user: User):
        self.user = user
        self.context = None
        self.session_id = None

    def get_completions(self, document: Document, complete_event):
        predictions = self.user.get_completions(session_id=self.session_id, events=self.context, query=document.text)
        for p in predictions:
            yield Completion(p, start_position=0)


class AutoComplete:
    def __init__(self, user, dialogue_dataset, interactive=True, unprompted=False, multiline=False):
        self.user = user
        self.context = []
        self.dialogue_dataset = dialogue_dataset
        self.unprompted = unprompted
        self.multiline = multiline
        self.interactive = interactive
        self.session = PromptSession(complete_in_thread=True,
                                     complete_while_typing=True,
                                     completer=TypeGenieCompleter(user=self.user))
        self.session.default_buffer.on_text_changed.add_handler(text_changed)

    def sample_context_and_response(self):
        while True:
            did = randint(0, len(self.dialogue_dataset)-1)
            dialogue_events: List[Event] = self.dialogue_dataset[did].events

            agent_uids = [idx for idx, u in enumerate(dialogue_events) if u.author == 'AGENT' and u.event == 'MESSAGE']
            if len(agent_uids) > 0:
                break

        split_idx = int(choice(agent_uids))
        return dialogue_events[:split_idx], dialogue_events[split_idx:]

    def interact(self):
        while True:
            try:
                printc('-' * 100, fore=F.WHITE)
                context, remaining = self.sample_context_and_response()
                self.session.completer.session_id = self.user.create_session()
                self.render_context(context)
                self.context = context
                for i in range(len(remaining)):
                    event = remaining[i]
                    if event.event != 'MESSAGE':
                        continue

                    role, response = event.author, event.value
                    if role == 'AGENT':
                        printc('\nAgent actually said: ' + response, fore=F.BLUE)
                        new_response = self.get_prediction()
                        if len(new_response) > 0:
                            response = new_response
                    else:
                        printc('\nUser: ' + response, fore=F.YELLOW)

                    if not self.interactive:
                        break
                    event._value = response
                    self.context.append(event)

                printc('-' * 100, fore=F.WHITE)
            except KeyboardInterrupt:
                self.context = []
                printc('\nChat reset!', F.RED)
                printc('\nPress cntr + C again to exit', F.RED)
                try:
                    time.sleep(1)
                except KeyboardInterrupt:
                    printc('\nExiting...', F.RED)
                    return

    def render_context(self, context: List[Event]):
        for event in context:
            if event.event == 'MESSAGE':
                if event.author == 'USER':
                    printc('\nUser: ' + event.value, fore=F.YELLOW)
                elif event.author == 'AGENT':
                    printc('\nAgent: ' + event.value, fore=F.GREEN)

    def get_prediction(self):
        self.session.completer.context = self.context
        text = self.session.prompt('Agent (with TypeGenie): ', pre_run=pre_run if self.unprompted else None,
                                   multiline=self.multiline)
        return text




import time
from datetime import datetime, timezone
from typing import List
from json import JSONDecodeError

from threading import Thread

import requests
from requests.auth import HTTPBasicAuth

BASE_URL = 'https://api.typegenie.net/api/v1'
SANDBOX_URL = 'https://staging-api.typegenie.net/api/v1'


class BearerAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r


class AutoRenewThread(Thread):
    def __init__(self, api):
        super().__init__()
        self.api = api
        self._running = False

    def run(self):
        while self._running:
            result = self.api.renew_token(inplace=True)
            expires_at = datetime.strptime(result['expires_at'], "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=timezone.utc)
            seconds_till_expiry = (expires_at - datetime.utcnow().replace(tzinfo=timezone.utc)).seconds
            renew_after = max(0, seconds_till_expiry - 100)
            time.sleep(renew_after)

    def start(self):
        if not self._running:
            self._running = True
            super().start()

    def stop(self):
        self._running = False


class API:
    COMMON_PREFIX = ''

    def __init__(self, sandbox=False):
        self._session = requests.Session()
        self._session.headers.update({
            'Content-type': 'application/json',
            'Accept': 'application/json'
        })
        self._sandbox = sandbox

    def _get_url(self, end_point):
        if self._sandbox:
            return f'{SANDBOX_URL}{self.COMMON_PREFIX}{end_point}'
        else:
            return f'{BASE_URL}{self.COMMON_PREFIX}{end_point}'

    def _request(self, func, url, **kwargs):
        try:
            resp = func(url, **kwargs)
            resp.raise_for_status()
            r = resp.json()
            return r['result']
        except requests.exceptions.HTTPError as err:
            try:
                r = resp.json()
                err.args = list(err.args) + [f'Reason: {r["error"]}']
                raise err
            except JSONDecodeError:
                raise err

    def _get(self, endpoint):
        url = self._get_url(endpoint)
        return self._request(self._session.get, url=url)

    def _delete(self, endpoint):
        url = self._get_url(endpoint)
        return self._request(self._session.delete, url=url)

    def _post(self, endpoint, json):
        url = self._get_url(endpoint)
        return self._request(self._session.post, url=url, json=json)

    def _put(self, endpoint, json):
        url = self._get_url(endpoint)
        return self._request(self._session.put, url=url, json=json)


class AccountAPI(API):
    COMMON_PREFIX = '/account'
    DEPLOYMENT_SUFFIX = '/deployment'

    def __init__(self, username, password, sandbox=False):
        super().__init__(sandbox=sandbox)
        self._session.auth = HTTPBasicAuth(username=username, password=password)

    def list_deployments(self):
        return self._get(self.DEPLOYMENT_SUFFIX)

    def create_deployment(self, deployment_id, metadata={}):
        assert isinstance(deployment_id, str)
        payload = {'id': deployment_id, 'metadata': metadata}
        return self._post(self.DEPLOYMENT_SUFFIX, json=payload)

    def update_deployement(self, deployment_id, metadata):
        assert isinstance(deployment_id, str)
        payload = {'metadata': metadata}
        return self._put(f'{self.DEPLOYMENT_SUFFIX}/{deployment_id}', json=payload)

    def delete_deployment(self, deployment_id):
        assert isinstance(deployment_id, str)
        return self._delete(f'{self.DEPLOYMENT_SUFFIX}/{deployment_id}')

    def get_deployment(self, deployment_id):
        assert isinstance(deployment_id, str)
        return self._get(f'{self.DEPLOYMENT_SUFFIX}/{deployment_id}')

    def get_deployment_access_token(self, deployment_id):
        assert isinstance(deployment_id, str)
        return self._get(f'{self.DEPLOYMENT_SUFFIX}/{deployment_id}/token')


class DeploymentAPI(API):
    COMMON_PREFIX = '/deployment'
    USER_SUFFIX = '/user'
    MODEL_SUFFIX = '/model'
    DATASET_SUFFIX = '/dataset'

    def __init__(self, token, sandbox=False):
        super().__init__(sandbox=sandbox)
        self._session.auth = BearerAuth(token=token)
        self._auto_renew = AutoRenewThread(api=self)

    def enable_auto_renew(self):
        self._auto_renew.start()

    def disable_auto_renew(self):
        self._auto_renew.stop()

    def info(self):
        return self._get('')

    def renew_token(self, inplace=False):
        result = self._get('/renew')
        if inplace:
            self._session.auth = BearerAuth(token=result['token'])
        return result

    # Dataset Endpoints
    def create_dataset(self, dataset_id, metadata={}):
        assert isinstance(dataset_id, str)
        payload = {'id': dataset_id, 'metadata': metadata}
        return self._post(self.DATASET_SUFFIX, json=payload)

    def list_datasets(self):
        return self._get(self.DATASET_SUFFIX)

    def get_dataset(self, dataset_id):
        assert isinstance(dataset_id, str)
        return self._get(f'{self.DATASET_SUFFIX}/{dataset_id}')

    def download_dataset(self, dataset_id):
        assert isinstance(dataset_id, str)
        return self._get(f'{self.DATASET_SUFFIX}/{dataset_id}/download')

    def update_dataset(self, dataset_id, metadata):
        assert isinstance(dataset_id, str)
        payload = {'metadata': metadata}
        return self._put(f'{self.DATASET_SUFFIX}/{dataset_id}', json=payload)

    def delete_dataset(self, dataset_id):
        assert isinstance(dataset_id, str)
        return self._delete(f'{self.DATASET_SUFFIX}/{dataset_id}')

    def upload_dialogues(self, dataset_id, dialogues):
        assert isinstance(dataset_id, str)
        payload = {'dialogues': dialogues}
        return self._post(f'{self.DATASET_SUFFIX}/{dataset_id}', json=payload)

    # Model Endpoints
    def list_model_configs(self):
        return self._get(f'{self.MODEL_SUFFIX}/configs')

    def deploy_model(self, dataset_ids: list, model_config_id):
        payload = {'dataset_ids': dataset_ids, 'configuration_id': model_config_id}
        return self._post(self.MODEL_SUFFIX, json=payload)

    def get_deployment_history(self):
        return self._get(self.MODEL_SUFFIX)

    def undeploy_model(self):
        return self._delete(self.MODEL_SUFFIX)

    # Agent Endpoints
    def add_user(self, user_id, metadata={}):
        assert isinstance(user_id, str)
        payload = {'id': user_id, 'metadata': metadata}
        return self._post(self.USER_SUFFIX, json=payload)

    def list_users(self):
        return self._get(f'{self.USER_SUFFIX}')

    def get_user(self, user_id):
        assert isinstance(user_id, str)
        return self._get(f'{self.USER_SUFFIX}/{user_id}')

    def update_user(self, user_id, metadata):
        assert isinstance(user_id, str)
        payload = {'metadata': metadata}
        return self._put(f'{self.USER_SUFFIX}/{user_id}', json=payload)

    def delete_user(self, user_id):
        assert isinstance(user_id, str)
        return self._delete(f'{self.USER_SUFFIX}/{user_id}')

    def get_user_access_token(self, user_id):
        assert isinstance(user_id, str)
        return self._get(f'{self.USER_SUFFIX}/{user_id}/token')


class UserAPI(API):
    COMMON_PREFIX = '/user'
    SESSION_SUFFIX = '/session'

    def __init__(self, token, sandbox=False):
        super().__init__(sandbox=sandbox)
        self._session.auth = BearerAuth(token=token)
        self._auto_renew = AutoRenewThread(api=self)

    def enable_auto_renew(self):
        self._auto_renew.start()

    def disable_auto_renew(self):
        self._auto_renew.stop()

    def info(self):
        return self._get('')

    def renew_token(self, inplace=False):
        result = self._get('/renew')
        if inplace:
            self._session.auth = BearerAuth(token=result['token'])
        return result

    def create_session(self):
        return self._post(self.SESSION_SUFFIX, json={})

    def get_completions(self, session_id, events: List[dict], query: str):
        assert isinstance(session_id, str)
        payload = {'query': query, 'events': events}
        return self._post(f'{self.SESSION_SUFFIX}/{session_id}/completions', json=payload)



import json
from pathlib import Path
import requests
from .lib import Deployment, Dialogue, authenticator
from .autocomplete import AutoComplete
import click
from box import Box


@click.command()
@click.option('-u', '--username', type=str, required=False, help="Account Username", default=None)
@click.option('-p', '--password', type=str, required=False, help="Account Password", default=None)
@click.option('-dpi', '--deployment_id', type=str, required=True, help="Deployment ID")
@click.option('-t', '--token', type=str, required=False, help="Deployment access token.")
@click.option('-dsi', '--dataset_id', type=str, required=True, help="DatasetID to be used for testing.")
@click.option('-ui', '--user-id', type=str, required=True, help="User ID within the Deployment")
@click.option('--production', is_flag=True, default=False, help="By default, sandbox environment is used. Set this "
                                                                "flag to use production environment")
@click.option('--cache-dir', type=click.Path(exists=True, file_okay=False, dir_okay=True), default='/tmp',
              help="Path where dataset is downloaded")
@click.option('-n', '--num-dialogues', default=500, type=int, help="Number of dialogues to load")
@click.option('--interactive', is_flag=True, default=False, help="Set to continue interaction")
@click.option('--unprompted', is_flag=True, default=False, help="Show completions even when unprompted")
@click.option('--multiline', is_flag=True, default=False, help="Set to allow multiline completions")
def main(**params):
    params = Box(params)

    if not params.production:
        authenticator.enable_sandbox()

    if params.token is None:
        assert_str = "Either specify deployment `token` or account `username` and `password`"
        assert all(p is not None for p in [params.username, params.password]), assert_str
        authenticator.authenticate_account(username=params.username, password=params.password)
    else:
        authenticator.authenticate_deployment(token=params.token)

    authenticator.enable_auto_fallback()
    authenticator.enable_auto_renew()
    deployment = Deployment.get(deployment_id=params.deployment_id)
    print(f'Loaded Deployment: {deployment}')

    available_datasets = deployment.datasets()
    available_users = deployment.users()
    assert any(
        d.id == params.dataset_id for d in available_datasets), f"{params.dataset_id} not found in the deployment!"
    assert any(
        u.id == params.user_id for u in available_users), f"{params.user_id} not found in the deployment!"
    dataset = [d for d in available_datasets if d.id == params.dataset_id][0]
    user = [u for u in available_users if u.id == params.user_id][0]

    print(f'Downloading {dataset.num_dialogues} to {params.cache_dir}...')

    download_dir = Path(params.cache_dir) / dataset.id

    download_dir.mkdir(exist_ok=True, parents=True)
    download_links = dataset.get_download_links()

    dialogues = []
    for link in sorted(download_links, key=lambda x: x['url']):
        file_name = link['url'].split('?')[0].split('/')[-1]
        output_file = download_dir / file_name
        if not output_file.exists():
            data = requests.get(link['url']).json()
            json.dump(data, output_file.open('w'), indent=4)
        else:
            data = json.load(output_file.open('r'))

        for d in data:
            if len(dialogues) < params.num_dialogues:
                dialogues.append(Dialogue.from_dict(d))
        else:
            if len(dialogues) >= params.num_dialogues:
                break

    autocomplete = AutoComplete(user=user,
                                dialogue_dataset=dialogues,
                                unprompted=params.unprompted,
                                multiline=params.multiline,
                                interactive=params.interactive)

    autocomplete.interact()


if __name__ == '__main__':
    main()
